# SPDX-License-Identifier: MIT

from __future__ import annotations

import asyncio
import time
from typing import TYPE_CHECKING, Callable, Dict, Iterable, List, Literal, Optional, Sequence, Union

from .abc import GuildChannel, Messageable
from .enums import ChannelType, ThreadArchiveDuration, try_enum, try_enum_to_int
from .errors import ClientException
from .flags import ChannelFlags
from .mixins import Hashable
from .partial_emoji import PartialEmoji, _EmojiTag
from .permissions import Permissions
from .utils import MISSING, _get_as_snowflake, _unique, parse_time, snowflake_time

__all__ = (
    "Thread",
    "ThreadMember",
    "ForumTag",
)

if TYPE_CHECKING:
    import datetime

    from typing_extensions import Self

    from .abc import Snowflake, SnowflakeTime
    from .channel import CategoryChannel, ForumChannel, MediaChannel, TextChannel
    from .emoji import Emoji
    from .guild import Guild
    from .member import Member
    from .message import Message, PartialMessage
    from .role import Role
    from .state import ConnectionState
    from .types.snowflake import SnowflakeList
    from .types.threads import (
        ForumTag as ForumTagPayload,
        PartialForumTag as PartialForumTagPayload,
        Thread as ThreadPayload,
        ThreadArchiveDurationLiteral,
        ThreadMember as ThreadMemberPayload,
        ThreadMetadata,
    )

    AnyThreadArchiveDuration = Union[ThreadArchiveDuration, ThreadArchiveDurationLiteral]

    ThreadType = Literal[
        ChannelType.news_thread, ChannelType.public_thread, ChannelType.private_thread
    ]


class Thread(Messageable, Hashable):
    """Represents a Discord thread.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two threads are equal.

        .. describe:: x != y

            Checks if two threads are not equal.

        .. describe:: hash(x)

            Returns the thread's hash.

        .. describe:: str(x)

            Returns the thread's name.

    .. versionadded:: 2.0

    Attributes
    ----------
    name: :class:`str`
        The thread name.
    guild: :class:`Guild`
        The guild the thread belongs to.
    id: :class:`int`
        The thread ID.
    parent_id: :class:`int`
        The parent :class:`TextChannel`, :class:`ForumChannel`, or :class:`MediaChannel` ID this thread belongs to.
    owner_id: Optional[:class:`int`]
        The user's ID that created this thread.
    last_message_id: Optional[:class:`int`]
        The last message ID of the message sent to this thread. It may
        *not* point to an existing or valid message.
    slowmode_delay: :class:`int`
        The number of seconds a member must wait between sending messages
        in this thread. A value of `0` denotes that it is disabled.
        Bots, and users with :attr:`~Permissions.manage_channels` or
        :attr:`~Permissions.manage_messages`, bypass slowmode.
    message_count: Optional[:class:`int`]
        An approximate number of messages in this thread.

        .. note::

            If the thread was created before July 1, 2022, this could be inaccurate.
    member_count: Optional[:class:`int`]
        An approximate number of members in this thread. This caps at 50.
    total_message_sent: Optional[:class:`int`]
        The total number of messages sent in the thread, including deleted messages.

        .. versionadded:: 2.6

        .. note::

            If the thread was created before July 1, 2022, this could be inaccurate.
    me: Optional[:class:`ThreadMember`]
        A thread member representing yourself, if you've joined the thread.
        This could not be available.
    archived: :class:`bool`
        Whether the thread is archived.
    locked: :class:`bool`
        Whether the thread is locked.
    invitable: :class:`bool`
        Whether non-moderators can add other non-moderators to this thread.
        This is always ``True`` for public threads.
    auto_archive_duration: :class:`int`
        The duration in minutes until the thread is automatically archived due to inactivity.
        Usually a value of 60, 1440, 4320 and 10080.
    archive_timestamp: :class:`datetime.datetime`
        An aware timestamp of when the thread's archived status was last updated in UTC.
    create_timestamp: Optional[:class:`datetime.datetime`]
        An aware timestamp of when the thread was created in UTC.
        This is only available for threads created after 2022-01-09.

        .. versionadded:: 2.4

    last_pin_timestamp: Optional[:class:`datetime.datetime`]
        The time the most recent message was pinned, or ``None`` if no message is currently pinned.

        .. versionadded:: 2.5
    """

    __slots__ = (
        "name",
        "id",
        "guild",
        "owner_id",
        "parent_id",
        "last_message_id",
        "message_count",
        "total_message_sent",
        "member_count",
        "slowmode_delay",
        "me",
        "locked",
        "archived",
        "invitable",
        "auto_archive_duration",
        "archive_timestamp",
        "create_timestamp",
        "last_pin_timestamp",
        "_flags",
        "_applied_tags",
        "_type",
        "_state",
        "_members",
    )

    def __init__(self, *, guild: Guild, state: ConnectionState, data: ThreadPayload) -> None:
        self._state: ConnectionState = state
        self.guild: Guild = guild
        self._members: Dict[int, ThreadMember] = {}
        self._from_data(data)

    async def _get_channel(self):
        return self

    def __repr__(self) -> str:
        return (
            f"<Thread id={self.id!r} name={self.name!r} parent={self.parent!r} "
            f"owner_id={self.owner_id!r} locked={self.locked!r} archived={self.archived!r} "
            f"flags={self.flags!r} applied_tags={self.applied_tags!r}>"
        )

    def __str__(self) -> str:
        return self.name

    def _from_data(self, data: ThreadPayload) -> None:
        self.id: int = int(data["id"])
        self.parent_id: int = int(data["parent_id"])
        self.owner_id: Optional[int] = _get_as_snowflake(data, "owner_id")
        self.name: str = data["name"]
        self._type: ThreadType = try_enum(ChannelType, data["type"])  # type: ignore
        self.last_message_id: Optional[int] = _get_as_snowflake(data, "last_message_id")
        self.slowmode_delay: int = data.get("rate_limit_per_user", 0)
        self.message_count: int = data.get("message_count") or 0
        self.total_message_sent: int = data.get("total_message_sent") or 0
        self.member_count: Optional[int] = data.get("member_count")
        self.last_pin_timestamp: Optional[datetime.datetime] = parse_time(
            data.get("last_pin_timestamp")
        )
        self._flags: int = data.get("flags", 0)
        self._applied_tags: List[int] = list(map(int, data.get("applied_tags", [])))
        self._unroll_metadata(data["thread_metadata"])

        try:
            member = data["member"]
        except KeyError:
            self.me = None
        else:
            self.me = ThreadMember(self, member)

    def _unroll_metadata(self, data: ThreadMetadata) -> None:
        self.archived: bool = data["archived"]
        self.auto_archive_duration: ThreadArchiveDurationLiteral = data["auto_archive_duration"]
        self.archive_timestamp: datetime.datetime = parse_time(data["archive_timestamp"])
        self.locked: bool = data.get("locked", False)
        self.invitable: bool = data.get("invitable", True)
        self.create_timestamp: Optional[datetime.datetime] = parse_time(
            data.get("create_timestamp")
        )

    def _update(self, data: ThreadPayload) -> None:
        try:
            self.name = data["name"]
        except KeyError:
            pass

        self.slowmode_delay = data.get("rate_limit_per_user", 0)
        self._flags = data.get("flags", 0)
        self._applied_tags = list(map(int, data.get("applied_tags", [])))

        try:
            self._unroll_metadata(data["thread_metadata"])
        except KeyError:
            pass

    @property
    def type(self) -> ThreadType:
        """:class:`ChannelType`: The channel's Discord type.

        This always returns :attr:`ChannelType.public_thread`,
        :attr:`ChannelType.private_thread`, or :attr:`ChannelType.news_thread`.
        """
        return self._type

    @property
    def parent(self) -> Optional[Union[TextChannel, ForumChannel, MediaChannel]]:
        """Optional[Union[:class:`TextChannel`, :class:`ForumChannel`, :class:`MediaChannel`]]: The parent channel this thread belongs to."""
        return self.guild.get_channel(self.parent_id)  # type: ignore

    @property
    def owner(self) -> Optional[Member]:
        """Optional[:class:`Member`]: The member this thread belongs to."""
        if self.owner_id is None:
            return None
        return self.guild.get_member(self.owner_id)

    @property
    def mention(self) -> str:
        """:class:`str`: The string that allows you to mention the thread."""
        return f"<#{self.id}>"

    @property
    def members(self) -> List[ThreadMember]:
        """List[:class:`ThreadMember`]: A list of thread members in this thread.

        This requires :attr:`Intents.members` to be properly filled. Most of the time however,
        this data is not provided by the gateway and a call to :meth:`fetch_members` is
        needed.
        """
        return list(self._members.values())

    @property
    def last_message(self) -> Optional[Message]:
        """Gets the last message in this channel from the cache.

        The message might not be valid or point to an existing message.

        .. admonition:: Reliable Fetching
            :class: helpful

            For a slightly more reliable method of fetching the
            last message, consider using either :meth:`history`
            or :meth:`fetch_message` with the :attr:`last_message_id`
            attribute.

        Returns
        -------
        Optional[:class:`Message`]
            The last message in this channel or ``None`` if not found.
        """
        return self._state._get_message(self.last_message_id) if self.last_message_id else None

    @property
    def category(self) -> Optional[CategoryChannel]:
        """The category channel the parent channel belongs to, if applicable.

        Raises
        ------
        ClientException
            The parent channel was not cached and returned ``None``.

        Returns
        -------
        Optional[:class:`CategoryChannel`]
            The parent channel's category.
        """
        parent = self.parent
        if parent is None:
            raise ClientException("Parent channel not found")
        return parent.category

    @property
    def category_id(self) -> Optional[int]:
        """The category channel ID the parent channel belongs to, if applicable.

        Raises
        ------
        ClientException
            The parent channel was not cached and returned ``None``.

        Returns
        -------
        Optional[:class:`int`]
            The parent channel's category ID.
        """
        parent = self.parent
        if parent is None:
            raise ClientException("Parent channel not found")
        return parent.category_id

    @property
    def created_at(self) -> datetime.datetime:
        """:class:`datetime.datetime`: Returns the thread's creation time in UTC.

        .. versionchanged:: 2.4
            If ``create_timestamp`` is provided by Discord, that will be used instead of the time in the ID.
        """
        return self.create_timestamp or snowflake_time(self.id)

    @property
    def flags(self) -> ChannelFlags:
        """:class:`.ChannelFlags`: The channel flags for this thread.

        .. versionadded:: 2.5
        """
        return ChannelFlags._from_value(self._flags)

    @property
    def jump_url(self) -> str:
        """A URL that can be used to jump to this thread.

        .. versionadded:: 2.4
        """
        return f"https://discord.com/channels/{self.guild.id}/{self.id}"

    def is_private(self) -> bool:
        """Whether the thread is a private thread.

        A private thread is only viewable by those that have been explicitly
        invited or have :attr:`~.Permissions.manage_threads` permissions.

        :return type: :class:`bool`
        """
        return self._type is ChannelType.private_thread

    def is_news(self) -> bool:
        """Whether the thread is a news thread.

        A news thread is a thread that has a parent that is a news channel,
        i.e. :meth:`.TextChannel.is_news` is ``True``.

        :return type: :class:`bool`
        """
        return self._type is ChannelType.news_thread

    def is_nsfw(self) -> bool:
        """Whether the thread is NSFW or not.

        An NSFW thread is a thread that has a parent that is an NSFW channel,
        i.e. :meth:`.TextChannel.is_nsfw` is ``True``.

        :return type: :class:`bool`
        """
        parent = self.parent
        return parent is not None and parent.is_nsfw()

    def is_pinned(self) -> bool:
        """Whether the thread is pinned in a :class:`ForumChannel` or :class:`MediaChannel`.

        Pinned threads are not affected by the auto archive duration.

        This is a shortcut to :attr:`self.flags.pinned <ChannelFlags.pinned>`.

        .. versionadded:: 2.5

        :return type: :class:`bool`
        """
        return self.flags.pinned

    @property
    def applied_tags(self) -> List[ForumTag]:
        """List[:class:`ForumTag`]: The tags currently applied to this thread.
        Only applicable to threads in channels of type :class:`ForumChannel` or :class:`MediaChannel`.

        .. versionadded:: 2.6
        """
        from .channel import ThreadOnlyGuildChannel  # cyclic import

        parent = self.parent
        if not isinstance(parent, ThreadOnlyGuildChannel):
            return []

        # threads may have tag IDs for tags that don't exist anymore
        return list(filter(None, map(parent._available_tags.get, self._applied_tags)))

    def permissions_for(
        self,
        obj: Union[Member, Role],
        /,
        *,
        ignore_timeout: bool = MISSING,
    ) -> Permissions:
        """Handles permission resolution for the :class:`~disnake.Member`
        or :class:`~disnake.Role`.

        Since threads do not have their own permissions, they inherit them
        from the parent channel.
        However, the permission context is different compared to a normal channel,
        so this method has different behavior than calling the parent's
        :attr:`GuildChannel.permissions_for <.abc.GuildChannel.permissions_for>`
        method directly.

        .. versionchanged:: 2.9
            Properly takes :attr:`Permissions.send_messages_in_threads`
            into consideration.

        Parameters
        ----------
        obj: Union[:class:`~disnake.Member`, :class:`~disnake.Role`]
            The object to resolve permissions for. This could be either
            a member or a role. If it's a role then member overwrites
            are not computed.
        ignore_timeout: :class:`bool`
            Whether or not to ignore the user's timeout.
            Defaults to ``False``.

            .. versionadded:: 2.4

            .. note::

                This only applies to :class:`~disnake.Member` objects.

            .. versionchanged:: 2.6

                The default was changed to ``False``.

        Raises
        ------
        ClientException
            The parent channel was not cached and returned ``None``

        Returns
        -------
        :class:`~disnake.Permissions`
            The resolved permissions for the member or role.
        """
        parent = self.parent
        if parent is None:
            raise ClientException("Parent channel not found")
        # n.b. GuildChannel is used here so implicit overrides are not applied based on send_messages
        base = GuildChannel.permissions_for(parent, obj, ignore_timeout=ignore_timeout)

        # if you can't send a message in a channel then you can't have certain
        # permissions as well
        if not base.send_messages_in_threads:
            base.send_tts_messages = False
            base.send_voice_messages = False
            base.send_polls = False
            base.mention_everyone = False
            base.embed_links = False
            base.attach_files = False

        # if you can't view a channel then you have no permissions there
        if not base.view_channel:
            denied = Permissions.all_channel()
            base.value &= ~denied.value

        return base

    async def delete_messages(self, messages: Iterable[Snowflake]) -> None:
        """|coro|

        Deletes a list of messages. This is similar to :meth:`Message.delete`
        except it bulk deletes multiple messages.

        As a special case, if the number of messages is 0, then nothing
        is done. If the number of messages is 1 then single message
        delete is done. If it's more than two, then bulk delete is used.

        You cannot bulk delete more than 100 messages or messages that
        are older than 14 days old.

        You must have the :attr:`~Permissions.manage_messages` permission to
        use this.

        Usable only by bot accounts.

        Parameters
        ----------
        messages: Iterable[:class:`abc.Snowflake`]
            An iterable of messages denoting which ones to bulk delete.

        Raises
        ------
        ClientException
            The number of messages to delete was more than 100.
        Forbidden
            You do not have proper permissions to delete the messages or
            you're not using a bot account.
        NotFound
            If single delete, then the message was already deleted.
        HTTPException
            Deleting the messages failed.
        """
        if not isinstance(messages, (list, tuple)):
            messages = list(messages)

        if len(messages) == 0:
            return  # do nothing

        if len(messages) == 1:
            message_id = messages[0].id
            await self._state.http.delete_message(self.id, message_id)
            return

        if len(messages) > 100:
            raise ClientException("Can only bulk delete messages up to 100 messages")

        message_ids: SnowflakeList = [m.id for m in messages]
        await self._state.http.delete_messages(self.id, message_ids)

    async def purge(
        self,
        *,
        limit: Optional[int] = 100,
        check: Callable[[Message], bool] = MISSING,
        before: Optional[SnowflakeTime] = None,
        after: Optional[SnowflakeTime] = None,
        around: Optional[SnowflakeTime] = None,
        oldest_first: Optional[bool] = False,
        bulk: bool = True,
    ) -> List[Message]:
        """|coro|

        Purges a list of messages that meet the criteria given by the predicate
        ``check``. If a ``check`` is not provided then all messages are deleted
        without discrimination.

        You must have the :attr:`~Permissions.manage_messages` permission to
        delete messages even if they are your own (unless you are a user
        account). The :attr:`~Permissions.read_message_history` permission is
        also needed to retrieve message history.

        Examples
        --------
        Deleting bot's messages ::

            def is_me(m):
                return m.author == client.user

            deleted = await thread.purge(limit=100, check=is_me)
            await thread.send(f'Deleted {len(deleted)} message(s)')

        Parameters
        ----------
        limit: Optional[:class:`int`]
            The number of messages to search through. This is not the number
            of messages that will be deleted, though it can be.
        check: Callable[[:class:`Message`], :class:`bool`]
            The function used to check if a message should be deleted.
            It must take a :class:`Message` as its sole parameter.
        before: Optional[Union[:class:`abc.Snowflake`, :class:`datetime.datetime`]]
            Same as ``before`` in :meth:`history`.
        after: Optional[Union[:class:`abc.Snowflake`, :class:`datetime.datetime`]]
            Same as ``after`` in :meth:`history`.
        around: Optional[Union[:class:`abc.Snowflake`, :class:`datetime.datetime`]]
            Same as ``around`` in :meth:`history`.
        oldest_first: Optional[:class:`bool`]
            Same as ``oldest_first`` in :meth:`history`.
        bulk: :class:`bool`
            If ``True``, use bulk delete. Setting this to ``False`` is useful for mass-deleting
            a bot's own messages without :attr:`Permissions.manage_messages`. When ``True``, will
            fall back to single delete if messages are older than two weeks.

        Raises
        ------
        Forbidden
            You do not have proper permissions to do the actions required.
        HTTPException
            Purging the messages failed.

        Returns
        -------
        List[:class:`.Message`]
            The list of messages that were deleted.
        """
        if check is MISSING:
            check = lambda m: True

        iterator = self.history(
            limit=limit, before=before, after=after, oldest_first=oldest_first, around=around
        )
        ret: List[Message] = []
        count = 0

        minimum_time = int((time.time() - 14 * 24 * 60 * 60) * 1000.0 - 1420070400000) << 22

        async def _single_delete_strategy(messages: Iterable[Message]) -> None:
            for m in messages:
                await m.delete()

        strategy = self.delete_messages if bulk else _single_delete_strategy

        async for message in iterator:
            if count == 100:
                to_delete = ret[-100:]
                await strategy(to_delete)
                count = 0
                await asyncio.sleep(1)

            if not check(message):
                continue

            if message.id < minimum_time:
                # older than 14 days old
                if count == 1:
                    await ret[-1].delete()
                elif count >= 2:
                    to_delete = ret[-count:]
                    await strategy(to_delete)

                count = 0
                strategy = _single_delete_strategy

            count += 1
            ret.append(message)

        # SOme messages remaining to poll
        if count >= 2:
            # more than 2 messages -> bulk delete
            to_delete = ret[-count:]
            await strategy(to_delete)
        elif count == 1:
            # delete a single message
            await ret[-1].delete()

        return ret

    async def edit(
        self,
        *,
        name: str = MISSING,
        archived: bool = MISSING,
        locked: bool = MISSING,
        invitable: bool = MISSING,
        slowmode_delay: int = MISSING,
        auto_archive_duration: AnyThreadArchiveDuration = MISSING,
        pinned: bool = MISSING,
        flags: ChannelFlags = MISSING,
        applied_tags: Sequence[Snowflake] = MISSING,
        reason: Optional[str] = None,
    ) -> Thread:
        """|coro|

        Edits the thread.

        Editing the thread requires :attr:`.Permissions.manage_threads`. The thread
        creator can also edit ``name``, ``archived``, ``auto_archive_duration`` and ``applied_tags``.
        Note that if the thread is locked then only those with :attr:`.Permissions.manage_threads`
        can unarchive a thread.

        The thread must be unarchived to be edited.

        Parameters
        ----------
        name: :class:`str`
            The new name of the thread.
        archived: :class:`bool`
            Whether to archive the thread or not.
        locked: :class:`bool`
            Whether to lock the thread or not.
        invitable: :class:`bool`
            Whether non-moderators can add other non-moderators to this thread.
            Only available for private threads.
        auto_archive_duration: Union[:class:`int`, :class:`ThreadArchiveDuration`]
            The new duration in minutes before a thread is automatically archived for inactivity.
            Must be one of ``60``, ``1440``, ``4320``, or ``10080``.
        slowmode_delay: :class:`int`
            Specifies the slowmode rate limit for users in this thread, in seconds.
            A value of ``0`` disables slowmode. The maximum value possible is ``21600``.
        pinned: :class:`bool`
            Whether to pin the thread or not. This is only available for threads created in a :class:`ForumChannel` or :class:`MediaChannel`.

            .. versionadded:: 2.5

        flags: :class:`ChannelFlags`
            The new channel flags to set for this thread. This will overwrite any existing flags set on this channel.
            If parameter ``pinned`` is provided, that will override the setting of :attr:`ChannelFlags.pinned`.

            .. versionadded:: 2.6

        applied_tags: Sequence[:class:`abc.Snowflake`]
            The new tags of the thread. Maximum of 5.
            Can also be used to reorder existing tags.

            This is only available for threads in a :class:`ForumChannel` or :class:`MediaChannel`.

            If :attr:`~ForumTag.moderated` tags are edited, :attr:`Permissions.manage_threads`
            permissions are required.

            See also :func:`add_tags` and :func:`remove_tags`.

            .. versionadded:: 2.6

        reason: Optional[:class:`str`]
            The reason for editing this thread. Shows up on the audit log.

            .. versionadded:: 2.5

        Raises
        ------
        Forbidden
            You do not have permissions to edit the thread.
        HTTPException
            Editing the thread failed.

        Returns
        -------
        :class:`Thread`
            The newly edited thread.
        """
        payload = {}
        if name is not MISSING:
            payload["name"] = str(name)
        if archived is not MISSING:
            payload["archived"] = archived
        if auto_archive_duration is not MISSING:
            payload["auto_archive_duration"] = try_enum_to_int(auto_archive_duration)
        if locked is not MISSING:
            payload["locked"] = locked
        if invitable is not MISSING:
            payload["invitable"] = invitable
        if slowmode_delay is not MISSING:
            payload["rate_limit_per_user"] = slowmode_delay

        if pinned is not MISSING:
            # create base flags if flags are provided, otherwise use the internal flags.
            flags = ChannelFlags._from_value(self._flags if flags is MISSING else flags.value)
            flags.pinned = pinned

        if flags is not MISSING:
            if not isinstance(flags, ChannelFlags):
                raise TypeError("flags field must be of type ChannelFlags")
            payload["flags"] = flags.value

        if applied_tags is not MISSING:
            payload["applied_tags"] = [t.id for t in applied_tags]

        data = await self._state.http.edit_channel(self.id, **payload, reason=reason)
        # The data payload will always be a Thread payload
        return Thread(data=data, state=self._state, guild=self.guild)  # type: ignore

    async def join(self) -> None:
        """|coro|

        Joins this thread.

        You must have :attr:`~Permissions.send_messages_in_threads` to join a thread.
        If the thread is private, :attr:`~Permissions.manage_threads` is also needed.

        Raises
        ------
        Forbidden
            You do not have permissions to join the thread.
        HTTPException
            Joining the thread failed.
        """
        await self._state.http.join_thread(self.id)

    async def leave(self) -> None:
        """|coro|

        Leaves this thread.

        Raises
        ------
        HTTPException
            Leaving the thread failed.
        """
        await self._state.http.leave_thread(self.id)

    async def add_user(self, user: Snowflake) -> None:
        """|coro|

        Adds a user to this thread.

        You must have :attr:`~.Permissions.send_messages` permission to add a user to a public thread.
        If the thread is private then :attr:`~.Permissions.send_messages` and either :attr:`~.Permissions.create_private_threads`
        or :attr:`~.Permissions.manage_messages` permissions
        is required to add a user to the thread.

        Parameters
        ----------
        user: :class:`abc.Snowflake`
            The user to add to the thread.

        Raises
        ------
        Forbidden
            You do not have permissions to add the user to the thread.
        HTTPException
            Adding the user to the thread failed.
        """
        await self._state.http.add_user_to_thread(self.id, user.id)

    async def remove_user(self, user: Snowflake) -> None:
        """|coro|

        Removes a user from this thread.

        You must have :attr:`~Permissions.manage_threads` or be the creator of the thread to remove a user.

        Parameters
        ----------
        user: :class:`abc.Snowflake`
            The user to remove from the thread.

        Raises
        ------
        Forbidden
            You do not have permissions to remove the user from the thread.
        HTTPException
            Removing the user from the thread failed.
        """
        await self._state.http.remove_user_from_thread(self.id, user.id)

    async def fetch_member(self, member_id: int, /) -> ThreadMember:
        """|coro|

        Retrieves a single :class:`ThreadMember` from this thread.

        Parameters
        ----------
        member_id: :class:`int`
            The ID of the member to fetch.

        Raises
        ------
        NotFound
            The specified member was not found.
        HTTPException
            Retrieving the member failed.

        Returns
        -------
        :class:`ThreadMember`
            The thread member asked for.
        """
        member_data = await self._state.http.get_thread_member(self.id, member_id)
        return ThreadMember(parent=self, data=member_data)

    async def fetch_members(self) -> List[ThreadMember]:
        """|coro|

        Retrieves all :class:`ThreadMember` that are in this thread.

        This requires :attr:`Intents.members` to get information about members
        other than yourself.

        Raises
        ------
        HTTPException
            Retrieving the members failed.

        Returns
        -------
        List[:class:`ThreadMember`]
            All thread members in the thread.
        """
        members = await self._state.http.get_thread_members(self.id)
        return [ThreadMember(parent=self, data=data) for data in members]

    async def delete(self, *, reason: Optional[str] = None) -> None:
        """|coro|

        Deletes this thread.

        You must have :attr:`~Permissions.manage_threads` to delete threads.
        Alternatively, you may delete a thread if it's in a :class:`ForumChannel` or :class:`MediaChannel`,
        you are the thread creator, and there are no messages other than the initial message.

        Parameters
        ----------
        reason: Optional[:class:`str`]
            The reason for deleting this thread. Shows up on the audit log.

            .. versionadded:: 2.5

        Raises
        ------
        Forbidden
            You do not have permissions to delete this thread.
        HTTPException
            Deleting the thread failed.
        """
        await self._state.http.delete_channel(self.id, reason=reason)

    async def add_tags(self, *tags: Snowflake, reason: Optional[str] = None) -> None:
        """|coro|

        Adds the given tags to this thread, up to 5 in total.

        The thread must be in a :class:`ForumChannel` or :class:`MediaChannel`.

        Adding tags requires you to have :attr:`.Permissions.manage_threads` permissions,
        or be the owner of the thread.
        However, adding :attr:`~ForumTag.moderated` tags always requires :attr:`.Permissions.manage_threads` permissions.

        .. versionadded:: 2.6

        Parameters
        ----------
        *tags: :class:`abc.Snowflake`
            An argument list of :class:`abc.Snowflake` representing the :class:`ForumTag`\\s
            to add to the thread.
        reason: Optional[:class:`str`]
            The reason for editing this thread. Shows up on the audit log.

        Raises
        ------
        Forbidden
            You do not have permission to add these tags.
        HTTPException
            Editing the thread failed.
        """
        if not tags:
            return

        new_tags: List[int] = self._applied_tags.copy()
        new_tags.extend(t.id for t in tags)
        new_tags = _unique(new_tags)

        await self._state.http.edit_channel(self.id, applied_tags=new_tags, reason=reason)

    async def remove_tags(self, *tags: Snowflake, reason: Optional[str] = None) -> None:
        """|coro|

        Removes the given tags from this thread.

        The thread must be in a :class:`ForumChannel` or :class:`MediaChannel`.

        Removing tags requires you to have :attr:`.Permissions.manage_threads` permissions,
        or be the owner of the thread.
        However, removing :attr:`~ForumTag.moderated` tags always requires :attr:`.Permissions.manage_threads` permissions.

        .. versionadded:: 2.6

        Parameters
        ----------
        *tags: :class:`abc.Snowflake`
            An argument list of :class:`abc.Snowflake` representing the :class:`ForumTag`\\s
            to remove from the thread.
        reason: Optional[:class:`str`]
            The reason for editing this thread. Shows up on the audit log.

        Raises
        ------
        Forbidden
            You do not have permission to remove these tags.
        HTTPException
            Editing the thread failed.
        """
        if not tags:
            return

        to_remove = {t.id for t in tags}
        new_tags: List[int] = [tag_id for tag_id in self._applied_tags if tag_id not in to_remove]

        await self._state.http.edit_channel(self.id, applied_tags=new_tags, reason=reason)

    def get_partial_message(self, message_id: int, /) -> PartialMessage:
        """Creates a :class:`PartialMessage` from the message ID.

        This is useful if you want to work with a message and only have its ID without
        doing an unnecessary API call.

        .. versionadded:: 2.0

        Parameters
        ----------
        message_id: :class:`int`
            The message ID to create a partial message for.

        Returns
        -------
        :class:`PartialMessage`
            The partial message.
        """
        from .message import PartialMessage

        return PartialMessage(channel=self, id=message_id)

    def _add_member(self, member: ThreadMember) -> None:
        self._members[member.id] = member

    def _pop_member(self, member_id: int) -> Optional[ThreadMember]:
        return self._members.pop(member_id, None)


class ThreadMember(Hashable):
    """Represents a Discord thread member.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two thread members are equal.

        .. describe:: x != y

            Checks if two thread members are not equal.

        .. describe:: hash(x)

            Returns the thread member's hash.

        .. describe:: str(x)

            Returns the thread member's name.

    .. versionadded:: 2.0

    Attributes
    ----------
    id: :class:`int`
        The thread member's ID.
    thread_id: :class:`int`
        The thread's ID.
    joined_at: :class:`datetime.datetime`
        The time the member joined the thread in UTC.
    """

    __slots__ = (
        "id",
        "thread_id",
        "joined_at",
        "flags",
        "_state",
        "parent",
    )

    def __init__(self, parent: Thread, data: ThreadMemberPayload) -> None:
        self.parent = parent
        self._state = parent._state
        self._from_data(data)

    def __repr__(self) -> str:
        return (
            f"<ThreadMember id={self.id} thread_id={self.thread_id} joined_at={self.joined_at!r}>"
        )

    def _from_data(self, data: ThreadMemberPayload) -> None:
        try:
            self.id = int(data["user_id"])
        except KeyError as err:
            if (self_id := self._state.self_id) is None:
                raise AssertionError("self_id is None when updating our own ThreadMember.") from err
            self.id = self_id

        try:
            self.thread_id = int(data["id"])
        except KeyError:
            self.thread_id = self.parent.id

        self.joined_at = parse_time(data["join_timestamp"])
        self.flags = data["flags"]

    @property
    def thread(self) -> Thread:
        """:class:`Thread`: The thread this member belongs to."""
        return self.parent


class ForumTag(Hashable):
    """Represents a tag for threads in forum/media channels.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two tags are equal.

        .. describe:: x != y

            Checks if two tags are not equal.

        .. describe:: hash(x)

            Returns the tag's hash.

        .. describe:: str(x)

            Returns the tag's name.

    .. versionadded:: 2.6


    Examples
    --------
    Creating a new tag:

    .. code-block:: python3

        tags = forum.available_tags
        tags.append(ForumTag(name="cool new tag", moderated=True))
        await forum.edit(available_tags=tags)

    Editing an existing tag:

    .. code-block:: python3

        tags = []
        for tag in forum.available_tags:
            if tag.id == 1234:
                tag = tag.with_changes(name="whoa new name")
            tags.append(tag)
        await forum.edit(available_tags=tags)


    Attributes
    ----------
    id: :class:`int`
        The tag's ID. Note that if this tag was manually constructed,
        this will be ``0``.
    name: :class:`str`
        The tag's name.
    moderated: :class:`bool`
        Whether only moderators can add this tag to threads or remove it.
        Defaults to ``False``.
    emoji: Optional[Union[:class:`Emoji`, :class:`PartialEmoji`]]
        The emoji associated with this tag, if any.
        Due to a Discord limitation, this will have an empty
        :attr:`~PartialEmoji.name` if it is a custom :class:`PartialEmoji`.
    """

    __slots__ = ("id", "name", "moderated", "emoji")

    def __init__(
        self,
        *,
        name: str,
        emoji: Optional[Union[str, PartialEmoji, Emoji]] = None,
        moderated: bool = False,
    ) -> None:
        self.id: int = 0
        self.name: str = name
        self.moderated: bool = moderated

        self.emoji: Optional[Union[Emoji, PartialEmoji]] = None
        if emoji is None:
            self.emoji = None
        elif isinstance(emoji, str):
            self.emoji = PartialEmoji.from_str(emoji)
        elif isinstance(emoji, _EmojiTag):
            self.emoji = emoji
        else:
            raise TypeError("emoji must be None, a str, PartialEmoji, or Emoji instance.")

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return (
            f"<ForumTag id={self.id!r} name={self.name!r}"
            f" moderated={self.moderated!r} emoji={self.emoji!r}>"
        )

    @property
    def created_at(self) -> datetime.datetime:
        """:class:`datetime.datetime`: Returns the tag's creation time in UTC.

        .. versionadded:: 2.10
        """
        return snowflake_time(self.id)

    def to_dict(self) -> PartialForumTagPayload:
        emoji_name, emoji_id = PartialEmoji._emoji_to_name_id(self.emoji)
        data: PartialForumTagPayload = {
            "name": self.name,
            "emoji_id": emoji_id,
            "emoji_name": emoji_name,
            "moderated": self.moderated,
        }

        if self.id:
            data["id"] = self.id

        return data

    @classmethod
    def _from_data(cls, *, data: ForumTagPayload, state: ConnectionState) -> Self:
        emoji = state._get_emoji_from_fields(
            name=data.get("emoji_name"),
            id=_get_as_snowflake(data, "emoji_id"),
        )

        self = cls(
            name=data["name"],
            emoji=emoji,
            moderated=data["moderated"],
        )
        self.id = int(data["id"])

        return self

    def with_changes(
        self,
        *,
        name: str = MISSING,
        emoji: Optional[Union[str, Emoji, PartialEmoji]] = MISSING,
        moderated: bool = MISSING,
    ) -> Self:
        """Returns a new instance with the given changes applied,
        for easy use with :func:`ForumChannel.edit` or :func:`MediaChannel.edit`.
        All other fields will be kept intact.

        Returns
        -------
        :class:`ForumTag`
            The new tag instance.
        """
        new_self = self.__class__(
            name=self.name if name is MISSING else name,
            emoji=self.emoji if emoji is MISSING else emoji,
            moderated=self.moderated if moderated is MISSING else moderated,
        )
        new_self.id = self.id
        return new_self



# SPDX-License-Identifier: MIT

from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, List, Optional

from .enums import SubscriptionStatus, try_enum
from .mixins import Hashable
from .utils import parse_time, snowflake_time

if TYPE_CHECKING:
    from .state import ConnectionState
    from .types.subscription import Subscription as SubscriptionPayload
    from .user import User

__all__ = ("Subscription",)


class Subscription(Hashable):
    """Represents a subscription.

    This can only be retrieved using :meth:`SKU.subscriptions` or :meth:`SKU.fetch_subscription`,
    or provided by events (e.g. :func:`on_subscription_create`).

    .. warning::
        :class:`Subscription`\\s should not be used to grant perks. Use :class:`Entitlement`\\s as a way of determining whether a user should have access to a specific :class:`SKU`.

    .. note::
        Some subscriptions may have been canceled already; consider using :meth:`is_canceled` to check whether a given subscription was canceled.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two :class:`Subscription`\\s are equal.

        .. describe:: x != y

            Checks if two :class:`Subscription`\\s are not equal.

        .. describe:: hash(x)

            Returns the subscription's hash.

    .. versionadded:: 2.10

    Attributes
    ----------
    id: :class:`int`
        The subscription's ID.
    user_id: :class:`int`
        The ID of the user who is subscribed to the :attr:`sku_ids`.

        See also :attr:`user`.
    sku_ids: List[:class:`int`]
        The ID of the SKUs the user is subscribed to.
    renewal_sku_ids: List[:class:`int`]
        The IDs of the SKUs that will be renewed at the start of the new period.
    entitlement_ids: List[:class:`int`]
        The IDs of the entitlements the user has as part of this subscription.
    current_period_start: :class:`datetime.datetime`
        The time at which the current period for the given subscription started.
    current_period_end: :class:`datetime.datetime`
        The time at which the current period for the given subscription will end.
    status: :class:`SubscriptionStatus`
        The current status of the given subscription.
    canceled_at: Optional[:class:`datetime.datetime`]
        The time at which the subscription was canceled.

        See also :attr:`is_canceled`.
    """

    __slots__ = (
        "_state",
        "id",
        "user_id",
        "sku_ids",
        "entitlement_ids",
        "renewal_sku_ids",
        "current_period_start",
        "current_period_end",
        "status",
        "canceled_at",
    )

    def __init__(self, *, data: SubscriptionPayload, state: ConnectionState) -> None:
        self._state: ConnectionState = state

        self.id: int = int(data["id"])
        self.user_id: int = int(data["user_id"])
        self.sku_ids: List[int] = list(map(int, data["sku_ids"]))
        self.entitlement_ids: List[int] = list(map(int, data["entitlement_ids"]))
        self.renewal_sku_ids: Optional[List[int]] = (
            list(map(int, renewal_sku_ids))
            if (renewal_sku_ids := data.get("renewal_sku_ids")) is not None
            else None
        )
        self.current_period_start: datetime.datetime = parse_time(data["current_period_start"])
        self.current_period_end: datetime.datetime = parse_time(data["current_period_end"])
        self.status: SubscriptionStatus = try_enum(SubscriptionStatus, data["status"])
        self.canceled_at: Optional[datetime.datetime] = parse_time(data["canceled_at"])

    @property
    def created_at(self) -> datetime.datetime:
        """:class:`datetime.datetime`: Returns the subscription's creation time in UTC."""
        return snowflake_time(self.id)

    @property
    def user(self) -> Optional[User]:
        """Optional[:class:`User`]: The user who is subscribed to the :attr:`sku_ids`.

        Requires the user to be cached.
        See also :attr:`user_id`.
        """
        return self._state.get_user(self.user_id)

    @property
    def is_canceled(self) -> bool:
        """:class:`bool`: Whether the subscription was canceled,
        based on :attr:`canceled_at`.
        """
        return self.canceled_at is not None



# SPDX-License-Identifier: MIT

from typing import Literal, Optional, TypedDict

from typing_extensions import NotRequired

from .member import Member
from .snowflake import Snowflake
from .user import User

GuildScheduledEventPrivacyLevel = Literal[2]
GuildScheduledEventStatus = Literal[1, 2, 3, 4]
GuildScheduledEventEntityType = Literal[1, 2, 3]


class GuildScheduledEventUser(TypedDict):
    guild_scheduled_event_id: Snowflake
    user: User
    member: NotRequired[Member]


class GuildScheduledEventEntityMetadata(TypedDict, total=False):
    location: str


class GuildScheduledEvent(TypedDict):
    id: Snowflake
    guild_id: Snowflake
    channel_id: Optional[Snowflake]
    creator_id: NotRequired[Optional[Snowflake]]
    name: str
    description: NotRequired[Optional[str]]
    scheduled_start_time: str
    scheduled_end_time: Optional[str]
    privacy_level: GuildScheduledEventPrivacyLevel
    status: GuildScheduledEventStatus
    entity_type: GuildScheduledEventEntityType
    entity_id: Optional[Snowflake]
    entity_metadata: Optional[GuildScheduledEventEntityMetadata]
    creator: NotRequired[User]
    user_count: NotRequired[int]
    image: NotRequired[Optional[str]]



# SPDX-License-Identifier: MIT

import typing as t

from mypy.plugin import Plugin


# FIXME: properly deprecate this in the future
class DisnakePlugin(Plugin):
    """Custom mypy plugin; no-op as of version 2.9."""


def plugin(version: str) -> t.Type[Plugin]:
    return DisnakePlugin



# SPDX-License-Identifier: MIT

from __future__ import annotations

import asyncio
import inspect
import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import disnake
import disnake.utils
from disnake.enums import Event

from ._types import _BaseCommand
from .base_core import InvokableApplicationCommand
from .ctx_menus_core import InvokableMessageCommand, InvokableUserCommand
from .slash_core import InvokableSlashCommand

if TYPE_CHECKING:
    from typing_extensions import Self

    from disnake.interactions import ApplicationCommandInteraction

    from ._types import MaybeCoro
    from .bot import AutoShardedBot, AutoShardedInteractionBot, Bot, InteractionBot
    from .context import Context
    from .core import Command

    AnyBot = Union[Bot, AutoShardedBot, InteractionBot, AutoShardedInteractionBot]


__all__ = (
    "CogMeta",
    "Cog",
)

FuncT = TypeVar("FuncT", bound=Callable[..., Any])

MISSING: Any = disnake.utils.MISSING
_log = logging.getLogger(__name__)


def _cog_special_method(func: FuncT) -> FuncT:
    func.__cog_special_method__ = None
    return func


class CogMeta(type):
    """A metaclass for defining a cog.

    Note that you should probably not use this directly. It is exposed
    purely for documentation purposes along with making custom metaclasses to intermix
    with other metaclasses such as the :class:`abc.ABCMeta` metaclass.

    For example, to create an abstract cog mixin class, the following would be done.

    .. code-block:: python3

        import abc

        class CogABCMeta(commands.CogMeta, abc.ABCMeta):
            pass

        class SomeMixin(metaclass=abc.ABCMeta):
            pass

        class SomeCogMixin(SomeMixin, commands.Cog, metaclass=CogABCMeta):
            pass

    .. note::

        When passing an attribute of a metaclass that is documented below, note
        that you must pass it as a keyword-only argument to the class creation
        like the following example:

        .. code-block:: python3

            class MyCog(commands.Cog, name='My Cog'):
                pass

    Attributes
    ----------
    name: :class:`str`
        The cog name. By default, it is the name of the class with no modification.
    description: :class:`str`
        The cog description. By default, it is the cleaned docstring of the class.

        .. versionadded:: 1.6

    command_attrs: Dict[:class:`str`, Any]
        A list of attributes to apply to every command inside this cog. The dictionary
        is passed into the :class:`Command` options at ``__init__``.
        If you specify attributes inside the command attribute in the class, it will
        override the one specified inside this attribute. For example:

        .. code-block:: python3

            class MyCog(commands.Cog, command_attrs=dict(hidden=True)):
                @commands.command()
                async def foo(self, ctx):
                    pass # hidden -> True

                @commands.command(hidden=False)
                async def bar(self, ctx):
                    pass # hidden -> False

    slash_command_attrs: Dict[:class:`str`, Any]
        A list of attributes to apply to every slash command inside this cog. The dictionary
        is passed into the options of every :class:`InvokableSlashCommand` at ``__init__``.
        Usage of this kwarg is otherwise the same as with ``command_attrs``.

        .. note:: This does not apply to instances of :class:`SubCommand` or :class:`SubCommandGroup`.

        .. versionadded:: 2.5

    user_command_attrs: Dict[:class:`str`, Any]
        A list of attributes to apply to every user command inside this cog. The dictionary
        is passed into the options of every :class:`InvokableUserCommand` at ``__init__``.
        Usage of this kwarg is otherwise the same as with ``command_attrs``.

        .. versionadded:: 2.5

    message_command_attrs: Dict[:class:`str`, Any]
        A list of attributes to apply to every message command inside this cog. The dictionary
        is passed into the options of every :class:`InvokableMessageCommand` at ``__init__``.
        Usage of this kwarg is otherwise the same as with ``command_attrs``.

        .. versionadded:: 2.5
    """

    __cog_name__: str
    __cog_settings__: Dict[str, Any]
    __cog_slash_settings__: Dict[str, Any]
    __cog_user_settings__: Dict[str, Any]
    __cog_message_settings__: Dict[str, Any]
    __cog_commands__: List[Command]
    __cog_app_commands__: List[InvokableApplicationCommand]
    __cog_listeners__: List[Tuple[str, str]]

    def __new__(cls: Type[CogMeta], *args: Any, **kwargs: Any) -> CogMeta:
        name, bases, attrs = args
        attrs["__cog_name__"] = kwargs.pop("name", name)
        attrs["__cog_settings__"] = kwargs.pop("command_attrs", {})
        attrs["__cog_slash_settings__"] = kwargs.pop("slash_command_attrs", {})
        attrs["__cog_user_settings__"] = kwargs.pop("user_command_attrs", {})
        attrs["__cog_message_settings__"] = kwargs.pop("message_command_attrs", {})

        description = kwargs.pop("description", None)
        if description is None:
            description = inspect.cleandoc(attrs.get("__doc__", ""))
        attrs["__cog_description__"] = description

        commands = {}
        app_commands = {}
        listeners = {}
        no_bot_cog = (
            "Commands or listeners must not start with cog_ or bot_ (in method {0.__name__}.{1})"
        )

        new_cls = super().__new__(cls, name, bases, attrs, **kwargs)
        for base in reversed(new_cls.__mro__):
            for elem, value in base.__dict__.items():
                if elem in commands:
                    del commands[elem]
                if elem in app_commands:
                    del app_commands[elem]
                if elem in listeners:
                    del listeners[elem]

                is_static_method = isinstance(value, staticmethod)
                if is_static_method:
                    value = value.__func__
                if isinstance(value, _BaseCommand):
                    if is_static_method:
                        raise TypeError(
                            f"Command in method {base}.{elem!r} must not be staticmethod."
                        )
                    if elem.startswith(("cog_", "bot_")):
                        raise TypeError(no_bot_cog.format(base, elem))
                    commands[elem] = value
                elif isinstance(value, InvokableApplicationCommand):
                    if is_static_method:
                        raise TypeError(
                            f"Application command in method {base}.{elem!r} must not be staticmethod."
                        )
                    if elem.startswith(("cog_", "bot_")):
                        raise TypeError(no_bot_cog.format(base, elem))
                    app_commands[elem] = value
                elif asyncio.iscoroutinefunction(value):
                    if hasattr(value, "__cog_listener__"):
                        if elem.startswith(("cog_", "bot_")):
                            raise TypeError(no_bot_cog.format(base, elem))
                        listeners[elem] = value

        new_cls.__cog_commands__ = list(commands.values())  # this will be copied in Cog.__new__
        new_cls.__cog_app_commands__ = list(app_commands.values())

        listeners_as_list = []
        for listener in listeners.values():
            for listener_name in listener.__cog_listener_names__:
                # I use __name__ instead of just storing the value so I can inject
                # the self attribute when the time comes to add them to the bot
                listeners_as_list.append((listener_name, listener.__name__))

        new_cls.__cog_listeners__ = listeners_as_list
        return new_cls

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args)

    @classmethod
    def qualified_name(cls) -> str:
        return cls.__cog_name__


class Cog(metaclass=CogMeta):
    """The base class that all cogs must inherit from.

    A cog is a collection of commands, listeners, and optional state to
    help group commands together. More information on them can be found on
    the :ref:`ext_commands_cogs` page.

    When inheriting from this class, the options shown in :class:`CogMeta`
    are equally valid here.
    """

    __cog_name__: ClassVar[str]
    __cog_settings__: ClassVar[Dict[str, Any]]
    __cog_commands__: ClassVar[List[Command]]
    __cog_app_commands__: ClassVar[List[InvokableApplicationCommand]]
    __cog_listeners__: ClassVar[List[Tuple[str, str]]]

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        # For issue 426, we need to store a copy of the command objects
        # since we modify them to inject `self` to them.
        # To do this, we need to interfere with the Cog creation process.
        self = super().__new__(cls)
        cmd_attrs = cls.__cog_settings__
        slash_cmd_attrs = cls.__cog_slash_settings__
        user_cmd_attrs = cls.__cog_user_settings__
        message_cmd_attrs = cls.__cog_message_settings__

        # Either update the command with the cog provided defaults or copy it.
        cog_app_commands: List[InvokableApplicationCommand] = []
        for c in cls.__cog_app_commands__:
            if isinstance(c, InvokableSlashCommand):
                c = c._update_copy(slash_cmd_attrs)
            elif isinstance(c, InvokableUserCommand):
                c = c._update_copy(user_cmd_attrs)
            elif isinstance(c, InvokableMessageCommand):
                c = c._update_copy(message_cmd_attrs)

            cog_app_commands.append(c)

        self.__cog_app_commands__ = tuple(cog_app_commands)  # type: ignore  # overriding ClassVar
        # Replace the old command objects with the new copies
        for app_command in self.__cog_app_commands__:
            setattr(self, app_command.callback.__name__, app_command)

        self.__cog_commands__ = tuple(c._update_copy(cmd_attrs) for c in cls.__cog_commands__)  # type: ignore  # overriding ClassVar

        lookup = {cmd.qualified_name: cmd for cmd in self.__cog_commands__}
        for command in self.__cog_commands__:
            setattr(self, command.callback.__name__, command)
            parent = command.parent
            if parent is not None:
                # Get the latest parent reference
                parent = lookup[parent.qualified_name]  # type: ignore

                # Update our parent's reference to our self
                parent.remove_command(command.name)  # type: ignore
                parent.add_command(command)  # type: ignore

        return self

    def get_commands(self) -> List[Command]:
        """Returns a list of commands the cog has.

        Returns
        -------
        List[:class:`.Command`]
            A :class:`list` of :class:`.Command`\\s that are
            defined inside this cog.

            .. note::

                This does not include subcommands.
        """
        return [c for c in self.__cog_commands__ if c.parent is None]

    def get_application_commands(self) -> List[InvokableApplicationCommand]:
        """Returns a list of application commands the cog has.

        Returns
        -------
        List[:class:`.InvokableApplicationCommand`]
            A :class:`list` of :class:`.InvokableApplicationCommand`\\s that are
            defined inside this cog.

            .. note::

                This does not include subcommands.
        """
        return list(self.__cog_app_commands__)

    def get_slash_commands(self) -> List[InvokableSlashCommand]:
        """Returns a list of slash commands the cog has.

        Returns
        -------
        List[:class:`.InvokableSlashCommand`]
            A :class:`list` of :class:`.InvokableSlashCommand`\\s that are
            defined inside this cog.

            .. note::

                This does not include subcommands.
        """
        return [c for c in self.__cog_app_commands__ if isinstance(c, InvokableSlashCommand)]

    def get_user_commands(self) -> List[InvokableUserCommand]:
        """Returns a list of user commands the cog has.

        Returns
        -------
        List[:class:`.InvokableUserCommand`]
            A :class:`list` of :class:`.InvokableUserCommand`\\s that are
            defined inside this cog.
        """
        return [c for c in self.__cog_app_commands__ if isinstance(c, InvokableUserCommand)]

    def get_message_commands(self) -> List[InvokableMessageCommand]:
        """Returns a list of message commands the cog has.

        Returns
        -------
        List[:class:`.InvokableMessageCommand`]
            A :class:`list` of :class:`.InvokableMessageCommand`\\s that are
            defined inside this cog.
        """
        return [c for c in self.__cog_app_commands__ if isinstance(c, InvokableMessageCommand)]

    @property
    def qualified_name(self) -> str:
        """:class:`str`: Returns the cog's specified name, not the class name."""
        return self.__cog_name__

    @property
    def description(self) -> str:
        """:class:`str`: Returns the cog's description, typically the cleaned docstring."""
        return self.__cog_description__

    @description.setter
    def description(self, description: str) -> None:
        self.__cog_description__ = description

    def walk_commands(self) -> Generator[Command, None, None]:
        """An iterator that recursively walks through this cog's commands and subcommands.

        Yields
        ------
        Union[:class:`.Command`, :class:`.Group`]
            A command or group from the cog.
        """
        from .core import GroupMixin

        for command in self.__cog_commands__:
            if command.parent is None:
                yield command
                if isinstance(command, GroupMixin):
                    yield from command.walk_commands()

    def get_listeners(self) -> List[Tuple[str, Callable[..., Any]]]:
        """Returns a :class:`list` of (name, function) listener pairs the cog has.

        Returns
        -------
        List[Tuple[:class:`str`, :ref:`coroutine <coroutine>`]]
            The listeners defined in this cog.
        """
        return [(name, getattr(self, method_name)) for name, method_name in self.__cog_listeners__]

    @classmethod
    def _get_overridden_method(cls, method: FuncT) -> Optional[FuncT]:
        """Return None if the method is not overridden. Otherwise returns the overridden method."""
        return getattr(method.__func__, "__cog_special_method__", method)

    @classmethod
    def listener(cls, name: Union[str, Event] = MISSING) -> Callable[[FuncT], FuncT]:
        """A decorator that marks a function as a listener.

        This is the cog equivalent of :meth:`.Bot.listen`.

        Parameters
        ----------
        name: Union[:class:`str`, :class:`.Event`]
            The name of the event being listened to. If not provided, it
            defaults to the function's name.

        Raises
        ------
        TypeError
            The function is not a coroutine function or a string or an :class:`.Event` enum member was not passed as
            the name.
        """
        if name is not MISSING and not isinstance(name, (str, Event)):
            raise TypeError(
                f"Cog.listener expected str or Enum but received {name.__class__.__name__!r} instead."
            )

        def decorator(func: FuncT) -> FuncT:
            actual = func
            if isinstance(actual, staticmethod):
                actual = actual.__func__
            if not asyncio.iscoroutinefunction(actual):
                raise TypeError("Listener function must be a coroutine function.")
            actual.__cog_listener__ = True
            to_assign = (
                actual.__name__
                if name is MISSING
                else (name if isinstance(name, str) else f"on_{name.value}")
            )
            try:
                actual.__cog_listener_names__.append(to_assign)
            except AttributeError:
                actual.__cog_listener_names__ = [to_assign]
            # we have to return `func` instead of `actual` because
            # we need the type to be `staticmethod` for the metaclass
            # to pick it up but the metaclass unfurls the function and
            # thus the assignments need to be on the actual function
            return func

        return decorator

    def has_error_handler(self) -> bool:
        """Whether the cog has an error handler.

        .. versionadded:: 1.7

        :return type: :class:`bool`
        """
        return not hasattr(self.cog_command_error.__func__, "__cog_special_method__")

    def has_slash_error_handler(self) -> bool:
        """Whether the cog has a slash command error handler.

        :return type: :class:`bool`
        """
        return not hasattr(self.cog_slash_command_error.__func__, "__cog_special_method__")

    def has_user_error_handler(self) -> bool:
        """Whether the cog has a user command error handler.

        :return type: :class:`bool`
        """
        return not hasattr(self.cog_user_command_error.__func__, "__cog_special_method__")

    def has_message_error_handler(self) -> bool:
        """Whether the cog has a message command error handler.

        :return type: :class:`bool`
        """
        return not hasattr(self.cog_message_command_error.__func__, "__cog_special_method__")

    @_cog_special_method
    async def cog_load(self) -> None:
        """A special method that is called as a task when the cog is added."""
        pass

    @_cog_special_method
    def cog_unload(self) -> None:
        """A special method that is called when the cog gets removed.

        This function **cannot** be a coroutine. It must be a regular
        function.

        Subclasses must replace this if they want special unloading behaviour.
        """
        pass

    @_cog_special_method
    def bot_check_once(self, ctx: Context) -> MaybeCoro[bool]:
        """A special method that registers as a :meth:`.Bot.check_once`
        check.

        This is for text commands only, and doesn't apply to application commands.

        This function **can** be a coroutine and must take a sole parameter,
        ``ctx``, to represent the :class:`.Context`.
        """
        return True

    @_cog_special_method
    def bot_check(self, ctx: Context) -> MaybeCoro[bool]:
        """A special method that registers as a :meth:`.Bot.check`
        check.

        This is for text commands only, and doesn't apply to application commands.

        This function **can** be a coroutine and must take a sole parameter,
        ``ctx``, to represent the :class:`.Context`.
        """
        return True

    @_cog_special_method
    def bot_slash_command_check_once(self, inter: ApplicationCommandInteraction) -> MaybeCoro[bool]:
        """A special method that registers as a :meth:`.Bot.slash_command_check_once`
        check.

        This function **can** be a coroutine and must take a sole parameter,
        ``inter``, to represent the :class:`.ApplicationCommandInteraction`.
        """
        return True

    @_cog_special_method
    def bot_slash_command_check(self, inter: ApplicationCommandInteraction) -> MaybeCoro[bool]:
        """A special method that registers as a :meth:`.Bot.slash_command_check`
        check.

        This function **can** be a coroutine and must take a sole parameter,
        ``inter``, to represent the :class:`.ApplicationCommandInteraction`.
        """
        return True

    @_cog_special_method
    def bot_user_command_check_once(self, inter: ApplicationCommandInteraction) -> MaybeCoro[bool]:
        """Similar to :meth:`.Bot.slash_command_check_once` but for user commands."""
        return True

    @_cog_special_method
    def bot_user_command_check(self, inter: ApplicationCommandInteraction) -> MaybeCoro[bool]:
        """Similar to :meth:`.Bot.slash_command_check` but for user commands."""
        return True

    @_cog_special_method
    def bot_message_command_check_once(
        self, inter: ApplicationCommandInteraction
    ) -> MaybeCoro[bool]:
        """Similar to :meth:`.Bot.slash_command_check_once` but for message commands."""
        return True

    @_cog_special_method
    def bot_message_command_check(self, inter: ApplicationCommandInteraction) -> MaybeCoro[bool]:
        """Similar to :meth:`.Bot.slash_command_check` but for message commands."""
        return True

    @_cog_special_method
    def cog_check(self, ctx: Context) -> MaybeCoro[bool]:
        """A special method that registers as a :func:`~.check`
        for every text command and subcommand in this cog.

        This is for text commands only, and doesn't apply to application commands.

        This function **can** be a coroutine and must take a sole parameter,
        ``ctx``, to represent the :class:`.Context`.
        """
        return True

    @_cog_special_method
    def cog_slash_command_check(self, inter: ApplicationCommandInteraction) -> MaybeCoro[bool]:
        """A special method that registers as a :func:`~.check`
        for every slash command and subcommand in this cog.

        This function **can** be a coroutine and must take a sole parameter,
        ``inter``, to represent the :class:`.ApplicationCommandInteraction`.
        """
        return True

    @_cog_special_method
    def cog_user_command_check(self, inter: ApplicationCommandInteraction) -> MaybeCoro[bool]:
        """Similar to :meth:`.Cog.cog_slash_command_check` but for user commands."""
        return True

    @_cog_special_method
    def cog_message_command_check(self, inter: ApplicationCommandInteraction) -> MaybeCoro[bool]:
        """Similar to :meth:`.Cog.cog_slash_command_check` but for message commands."""
        return True

    @_cog_special_method
    async def cog_command_error(self, ctx: Context, error: Exception) -> None:
        """A special method that is called whenever an error
        is dispatched inside this cog.

        This is for text commands only, and doesn't apply to application commands.

        This is similar to :func:`.on_command_error` except only applying
        to the commands inside this cog.

        This **must** be a coroutine.

        Parameters
        ----------
        ctx: :class:`.Context`
            The invocation context where the error happened.
        error: :class:`CommandError`
            The error that was raised.
        """
        pass

    @_cog_special_method
    async def cog_slash_command_error(
        self, inter: ApplicationCommandInteraction, error: Exception
    ) -> None:
        """A special method that is called whenever an error
        is dispatched inside this cog.

        This is similar to :func:`.on_slash_command_error` except only applying
        to the slash commands inside this cog.

        This **must** be a coroutine.

        Parameters
        ----------
        inter: :class:`.ApplicationCommandInteraction`
            The interaction where the error happened.
        error: :class:`CommandError`
            The error that was raised.
        """
        pass

    @_cog_special_method
    async def cog_user_command_error(
        self, inter: ApplicationCommandInteraction, error: Exception
    ) -> None:
        """Similar to :func:`cog_slash_command_error` but for user commands."""
        pass

    @_cog_special_method
    async def cog_message_command_error(
        self, inter: ApplicationCommandInteraction, error: Exception
    ) -> None:
        """Similar to :func:`cog_slash_command_error` but for message commands."""
        pass

    @_cog_special_method
    async def cog_before_invoke(self, ctx: Context) -> None:
        """A special method that acts as a cog local pre-invoke hook,
        similar to :meth:`.Command.before_invoke`.

        This is for text commands only, and doesn't apply to application commands.

        This **must** be a coroutine.

        Parameters
        ----------
        ctx: :class:`.Context`
            The invocation context.
        """
        pass

    @_cog_special_method
    async def cog_after_invoke(self, ctx: Context) -> None:
        """A special method that acts as a cog local post-invoke hook,
        similar to :meth:`.Command.after_invoke`.

        This is for text commands only, and doesn't apply to application commands.

        This **must** be a coroutine.

        Parameters
        ----------
        ctx: :class:`.Context`
            The invocation context.
        """
        pass

    @_cog_special_method
    async def cog_before_slash_command_invoke(self, inter: ApplicationCommandInteraction) -> None:
        """A special method that acts as a cog local pre-invoke hook.

        This is similar to :meth:`.Command.before_invoke` but for slash commands.

        This **must** be a coroutine.

        Parameters
        ----------
        inter: :class:`.ApplicationCommandInteraction`
            The interaction of the slash command.
        """
        pass

    @_cog_special_method
    async def cog_after_slash_command_invoke(self, inter: ApplicationCommandInteraction) -> None:
        """A special method that acts as a cog local post-invoke hook.

        This is similar to :meth:`.Command.after_invoke` but for slash commands.

        This **must** be a coroutine.

        Parameters
        ----------
        inter: :class:`.ApplicationCommandInteraction`
            The interaction of the slash command.
        """
        pass

    @_cog_special_method
    async def cog_before_user_command_invoke(self, inter: ApplicationCommandInteraction) -> None:
        """Similar to :meth:`cog_before_slash_command_invoke` but for user commands."""
        pass

    @_cog_special_method
    async def cog_after_user_command_invoke(self, inter: ApplicationCommandInteraction) -> None:
        """Similar to :meth:`cog_after_slash_command_invoke` but for user commands."""
        pass

    @_cog_special_method
    async def cog_before_message_command_invoke(self, inter: ApplicationCommandInteraction) -> None:
        """Similar to :meth:`cog_before_slash_command_invoke` but for message commands."""
        pass

    @_cog_special_method
    async def cog_after_message_command_invoke(self, inter: ApplicationCommandInteraction) -> None:
        """Similar to :meth:`cog_after_slash_command_invoke` but for message commands."""
        pass

    def _inject(self, bot: AnyBot) -> Self:
        from .bot import AutoShardedInteractionBot, InteractionBot

        cls = self.__class__

        if (
            isinstance(bot, (InteractionBot, AutoShardedInteractionBot))
            and len(self.__cog_commands__) > 0
        ):
            raise TypeError("@commands.command is not supported for interaction bots.")

        # realistically, the only thing that can cause loading errors
        # is essentially just the command loading, which raises if there are
        # duplicates. When this condition is met, we want to undo all what
        # we've added so far for some form of atomic loading.
        for index, command in enumerate(self.__cog_commands__):
            command.cog = self
            if command.parent is None:
                try:
                    bot.add_command(command)  # type: ignore
                except Exception:
                    # undo our additions
                    for to_undo in self.__cog_commands__[:index]:
                        if to_undo.parent is None:
                            bot.remove_command(to_undo.name)  # type: ignore
                    raise

        for index, command in enumerate(self.__cog_app_commands__):
            command.cog = self
            try:
                if isinstance(command, InvokableSlashCommand):
                    bot.add_slash_command(command)
                elif isinstance(command, InvokableUserCommand):
                    bot.add_user_command(command)
                elif isinstance(command, InvokableMessageCommand):
                    bot.add_message_command(command)
            except Exception:
                # undo our additions
                for to_undo in self.__cog_app_commands__[:index]:
                    if isinstance(to_undo, InvokableSlashCommand):
                        bot.remove_slash_command(to_undo.name)
                    elif isinstance(to_undo, InvokableUserCommand):
                        bot.remove_user_command(to_undo.name)
                    elif isinstance(to_undo, InvokableMessageCommand):
                        bot.remove_message_command(to_undo.name)
                raise

        if not hasattr(self.cog_load.__func__, "__cog_special_method__"):
            bot.loop.create_task(disnake.utils.maybe_coroutine(self.cog_load))

        # check if we're overriding the default
        if cls.bot_check is not Cog.bot_check:
            if isinstance(bot, (InteractionBot, AutoShardedInteractionBot)):
                raise TypeError("Cog.bot_check is not supported for interaction bots.")

            bot.add_check(self.bot_check)

        if cls.bot_check_once is not Cog.bot_check_once:
            if isinstance(bot, (InteractionBot, AutoShardedInteractionBot)):
                raise TypeError("Cog.bot_check_once is not supported for interaction bots.")

            bot.add_check(self.bot_check_once, call_once=True)

        # Add application command checks
        if cls.bot_slash_command_check is not Cog.bot_slash_command_check:
            bot.add_app_command_check(self.bot_slash_command_check, slash_commands=True)

        if cls.bot_user_command_check is not Cog.bot_user_command_check:
            bot.add_app_command_check(self.bot_user_command_check, user_commands=True)

        if cls.bot_message_command_check is not Cog.bot_message_command_check:
            bot.add_app_command_check(self.bot_message_command_check, message_commands=True)

        # Add app command one-off checks
        if cls.bot_slash_command_check_once is not Cog.bot_slash_command_check_once:
            bot.add_app_command_check(
                self.bot_slash_command_check_once,
                call_once=True,
                slash_commands=True,
            )

        if cls.bot_user_command_check_once is not Cog.bot_user_command_check_once:
            bot.add_app_command_check(
                self.bot_user_command_check_once, call_once=True, user_commands=True
            )

        if cls.bot_message_command_check_once is not Cog.bot_message_command_check_once:
            bot.add_app_command_check(
                self.bot_message_command_check_once,
                call_once=True,
                message_commands=True,
            )

        # while Bot.add_listener can raise if it's not a coroutine,
        # this precondition is already met by the listener decorator
        # already, thus this should never raise.
        # Outside of, memory errors and the like...
        for name, method_name in self.__cog_listeners__:
            bot.add_listener(getattr(self, method_name), name)

        try:
            if bot._command_sync_flags.sync_on_cog_actions:
                bot._schedule_delayed_command_sync()
        except NotImplementedError:
            pass

        return self

    def _eject(self, bot: AnyBot) -> None:
        cls = self.__class__

        try:
            for command in self.__cog_commands__:
                if command.parent is None:
                    bot.remove_command(command.name)  # type: ignore

            for app_command in self.__cog_app_commands__:
                if isinstance(app_command, InvokableSlashCommand):
                    bot.remove_slash_command(app_command.name)
                elif isinstance(app_command, InvokableUserCommand):
                    bot.remove_user_command(app_command.name)
                elif isinstance(app_command, InvokableMessageCommand):
                    bot.remove_message_command(app_command.name)

            for name, method_name in self.__cog_listeners__:
                bot.remove_listener(getattr(self, method_name), name)

            if cls.bot_check is not Cog.bot_check:
                bot.remove_check(self.bot_check)  # type: ignore

            if cls.bot_check_once is not Cog.bot_check_once:
                bot.remove_check(self.bot_check_once, call_once=True)  # type: ignore

            # Remove application command checks
            if cls.bot_slash_command_check is not Cog.bot_slash_command_check:
                bot.remove_app_command_check(self.bot_slash_command_check, slash_commands=True)

            if cls.bot_user_command_check is not Cog.bot_user_command_check:
                bot.remove_app_command_check(self.bot_user_command_check, user_commands=True)

            if cls.bot_message_command_check is not Cog.bot_message_command_check:
                bot.remove_app_command_check(self.bot_message_command_check, message_commands=True)

            # Remove app command one-off checks
            if cls.bot_slash_command_check_once is not Cog.bot_slash_command_check_once:
                bot.remove_app_command_check(
                    self.bot_slash_command_check_once,
                    call_once=True,
                    slash_commands=True,
                )

            if cls.bot_user_command_check_once is not Cog.bot_user_command_check_once:
                bot.remove_app_command_check(
                    self.bot_user_command_check_once,
                    call_once=True,
                    user_commands=True,
                )

            if cls.bot_message_command_check_once is not Cog.bot_message_command_check_once:
                bot.remove_app_command_check(
                    self.bot_message_command_check_once,
                    call_once=True,
                    message_commands=True,
                )

        finally:
            try:
                if bot._command_sync_flags.sync_on_cog_actions:
                    bot._schedule_delayed_command_sync()
            except NotImplementedError:
                pass
            try:
                self.cog_unload()
            except Exception as e:
                _log.error(
                    "An error occurred while unloading the %s cog.", self.qualified_name, exc_info=e
                )
